{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Import libraries"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import ee\n",
    "import geemap"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Create an interactive map"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "Map = geemap.Map(center=[40, -100], zoom=4)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Add Earth Engine Python script"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Add Earth Engine dataset\n",
    "image = ee.Image(\"USGS/SRTMGL1_003\")\n",
    "\n",
    "#*** Start of imports. If edited, may not auto-convert in the playground. ***#\n",
    "testArea =\n",
    "\n",
    "    # displayProperties: [\n",
    "      {\n",
    "        \"type\": \"rectangle\"\n",
    "      }\n",
    "    ] #\n",
    "    ee.Geometry.Polygon(\n",
    "        [[[-66.73156878460787, -8.662236005089952],\n",
    "          [-66.73156878460787, -8.916025640576244],\n",
    "          [-66.44867083538912, -8.916025640576244],\n",
    "          [-66.44867083538912, -8.662236005089952]]], None, False)\n",
    "#**** End of imports. If edited, may not auto-convert in the playground. ****#\n",
    "# Fusion Near Real-time (GUI)\n",
    "# Near real-time monitoring of forest disturbance by fusion of\n",
    "# multi-sensor data.  @author Xiaojing Tang (xjtang@bu.edu).\n",
    "\n",
    "# ---------------------------------------------------------------\n",
    "# Model Parameters:\n",
    "trainPeriod = ee.Dictionary({'start': '2017-01-01', 'end': '2020-01-01'})\n",
    "monitorPeriod = ee.Dictionary({'start': '2020-01-01', 'end': '2021-01-01'})\n",
    "fullPeriod = ee.Dictionary({\n",
    "     'start': trainPeriod.get('start'),\n",
    "     'end': monitorPeriod.get('end')})\n",
    "nrtParam = {'z': 2, 'n': 4, 'm': 5, 'maxZ': 10}\n",
    "optParam = {\n",
    "  'band': 'NDFI',\n",
    "  'minRMSE': 0.05,\n",
    "  'strikeOnly': False,\n",
    "  'vis': '{bands': ['SWIR1',  'NIR',  'Red'], 'min': 0, 'max': 5000}\n",
    "}\n",
    "radParam = {\n",
    "  'band': 'VV',\n",
    "  'minRMSE': 0.01,\n",
    "  'strikeOnly': True,\n",
    "  'vis': '{bands': ['VV',  'VH',  'ratio'], 'min': 10, 'max': 30}\n",
    "}\n",
    "mskVisParam = {'min': 0, 'max': 1, 'palette': ['blue', 'green']}\n",
    "altVisParam = {'min': 2020.4, 'max': 2021,\n",
    "                    'palette': ['FF0080', 'EC1280', 'DA2480', 'C83680', 'B64880', 'A35B80', '916D80',\n",
    "                              '7F7F80', '6D9180', '5BA380', '48B680', '36C880', '24DA80', '12EC80',\n",
    "                              '00FF80', '00EB89', '00D793', '00C49D', '00B0A7', '009CB0', '0089BA',\n",
    "                              '0075C4', '0062CE', '004ED7', '003AE1', '0027EB', '0013F5', '0000FF']}\n",
    "\n",
    "# ---------------------------------------------------------------\n",
    "# Imports and predefined variables:\n",
    "listener = 0\n",
    "ut = require('users/xjtang/fnrt:Lite/Utilities')\n",
    "hansen = ee.Image('UMD/hansen/global_forest_change_2020_v1_8').unmask()\n",
    "forestMask = hansen.select('treecover2000').gt(50) \\\n",
    "                  .add(hansen.select('gain')) \\\n",
    "                  .subtract(hansen.select('loss')) \\\n",
    "                  .add(hansen.select('lossyear').eq(20)) \\\n",
    "                  .gt(0)\n",
    "alerts = None\n",
    "ccd = ee.ImageCollection('projects/gee-book/assets/A3-5/ccd')\n",
    "\n",
    "# ---------------------------------------------------------------\n",
    "# Main functions:\n",
    "  # run and plot CCD result for a pixel\n",
    "def chartCCD(coords):\n",
    "  resetTSPanel()\n",
    "  sensor = ccdSelect.getValue()\n",
    "  pixel = ee.Geometry.Point([coords.lon, coords.lat])\n",
    "  tsPanel.add(ui.Label('Running CCD on ' + sensor + ' data...'))\n",
    "  addPixel(coords)\n",
    "  param = None\n",
    "  if (sensor == 'Sentinel-1') {\n",
    "    param = radParam\n",
    "  } else {\n",
    "    param = optParam\n",
    "  }\n",
    "  trainData = ut.getData(pixel, trainPeriod, sensor)\n",
    "  monitorData = ut.getData(pixel, monitorPeriod, sensor)\n",
    "  ccdModel = ee.Image(ee.Algorithms.If(\n",
    "                  ccd.filterMetadata('sensor', 'equals', sensor).first(),\n",
    "                  ccd.filterMetadata('sensor', 'equals', sensor).first(),\n",
    "                  ut.runCCD(trainData, trainPeriod, param.band)\n",
    "                ))\n",
    "  ccdTS = ut.getTimeSeries(trainData, monitorData, ccdModel, pixel, param.band, 0.1)\n",
    "  ccdTable = ut.transformToTable(ccdTS, ['dateString', 'train', 'monitor', 'fit'])\n",
    "  ccdTable.evaluate(function(t, e) {\n",
    "    chart = ut.createCCDChart(t, param.band, coords.lat, coords.lon)\n",
    "    chart.onClick(\n",
    "      def function(date):\n",
    "        if (date === None) {\n",
    "          removeLayer('_')\n",
    "        } else {\n",
    "          img = ee.Image(ut.getImage(pixel, date, sensor))\n",
    "          mapPanel.addLayer(img, param.vis, img.get('system:index').getInfo())\n",
    "          removeLayer('Clicked')\n",
    "          addPixel(coords)\n",
    "        }\n",
    "\n",
    "    )\n",
    "    tsPanel.clear()\n",
    "    tsPanel.add(chart)\n",
    "  })\n",
    "\n",
    "\n",
    "  # run near-real-time monitoring for one sensor for one pixel\n",
    "def runPixelSensorNRT(pixel, sensor):\n",
    "  trainData = ut.getData(pixel, trainPeriod, sensor)\n",
    "  monitorData = ut.getData(pixel, monitorPeriod, sensor)\n",
    "  param = None\n",
    "  if (sensor == 'Sentinel-1') {\n",
    "    param = radParam\n",
    "  } else {\n",
    "    param = optParam\n",
    "  }\n",
    "  ccdModel = ee.Image(ee.Algorithms.If(\n",
    "              ccd.filterMetadata('sensor', 'equals', sensor).first(),\n",
    "              ccd.filterMetadata('sensor', 'equals', sensor).first(),\n",
    "              ut.runCCD(trainData, trainPeriod, param.band)\n",
    "            ))\n",
    "  ccdTS = ut.getTimeSeries(trainData, monitorData, ccdModel, pixel, param.band)\n",
    "  Z = ee.FeatureCollection(ut.addPixelZScore(ccdTS, nrtParam.maxZ, param.minRMSE))\n",
    "  checked = ut.checkPixelZScore(Z, nrtParam.z)\n",
    "  if ((S2Check.getValue() | LSTCheck.getValue()) & param.strikeOnly) {\n",
    "    checked = checked.filterMetadata('Ball', 'equals', None)\n",
    "  }\n",
    "  return checked\n",
    "\n",
    "\n",
    "  # run and plot NRT for a pixel\n",
    "def chartNRT(coords):\n",
    "  resetTSPanel()\n",
    "  tsPanel.add(ui.Label('Running NRT...'))\n",
    "  pixel = ee.Geometry.Point([coords.lon, coords.lat])\n",
    "  addPixel(coords)\n",
    "  nrtTS = ee.FeatureCollection([])\n",
    "  if (S2Check.getValue()) {nrtTS = nrtTS.merge(runPixelSensorNRT(pixel, 'Sentinel-2'))}\n",
    "  if (S1Check.getValue()) {nrtTS = nrtTS.merge(runPixelSensorNRT(pixel, 'Sentinel-1'))}\n",
    "  if (LSTCheck.getValue()) {nrtTS = nrtTS.merge(runPixelSensorNRT(pixel, 'Landsat'))}\n",
    "  nrtTS = nrtTS.filterMetadata('x', 'not_equals', None)\n",
    "  if (nrtTS.size().gt(0)) {\n",
    "    nrtMonitor = ee.FeatureCollection(ut.monitorPixelChange(nrtTS.sort('fitTime'), nrtParam))\n",
    "    nrt = ut.transformToTable(nrtMonitor, ['dateString', 'Z_train', 'Ball', 'Strike', 'StrikeOut'])\n",
    "    nrt.evaluate(function(t, e) {\n",
    "      chart = ut.createNRTChart(t, coords.lat, coords.lon)\n",
    "      tsPanel.clear()\n",
    "      tsPanel.add(chart)\n",
    "    })\n",
    "  } else {\n",
    "    tsPanel.add(ui.Label('No sensor was selected.'))\n",
    "  }\n",
    "\n",
    "\n",
    "  # run near-real-time monitoring for one sensor for an area\n",
    "def runAreaSensorNRT(area, sensor):\n",
    "  param = None\n",
    "  if (sensor == 'Sentinel-1') {\n",
    "    param = radParam\n",
    "  } else {\n",
    "    param = optParam\n",
    "  }\n",
    "  monitorData = ut.getData(area, monitorPeriod, sensor)\n",
    "  ccdModel = ccd.filterMetadata('sensor', 'equals', sensor).first().updateMask(forestMask)\n",
    "  ccdImg = ut.getCCDImage(ccdModel, param.band)\n",
    "  synthetics = ut.addSynthetic(monitorData, ccdImg, param.band, sensor)\n",
    "  monitorRes = ut.getResiduals(synthetics, param.band)\n",
    "  return ut.getChangeScores(monitorRes, ccdImg.select('.*rmse'),\n",
    "                            synthetics.select(param.band).mean(),\n",
    "                            param.minRMSE, nrtParam.z, param.strikeOnly)\n",
    "\n",
    "\n",
    "  # run near real time\n",
    "def runNRT(area):\n",
    "  nrtTS = ee.ImageCollection([])\n",
    "  if (S2Check.getValue()) {nrtTS = nrtTS.merge(runAreaSensorNRT(area, 'Sentinel-2'))}\n",
    "  if (S1Check.getValue()) {nrtTS = nrtTS.merge(runAreaSensorNRT(area, 'Sentinel-1'))}\n",
    "  if (LSTCheck.getValue()) {nrtTS = nrtTS.merge(runAreaSensorNRT(area, 'Landsat'))}\n",
    "  if (nrtTS.size().gt(0)) {\n",
    "    nrtTS = nrtTS.sort('system:time_start')\n",
    "    return ut.monitorChange(nrtTS, nrtParam)\n",
    "  } else {\n",
    "    print('Nothing was selected.')\n",
    "    return ee.Image(0)\n",
    "  }\n",
    "\n",
    "\n",
    "  # add pixel location\n",
    "def addPixel(coords):\n",
    "  pSize = 0.000135\n",
    "  pixel = ee.Geometry.Rectangle([coords.lon - pSize, coords.lat - pSize,\n",
    "                                      coords.lon + pSize, coords.lat + pSize])\n",
    "  mapPanel.addLayer(pixel, {'color': '0000FF'}, 'Clicked')\n",
    "\n",
    "\n",
    "  # remove a layer\n",
    "def removeLayer(name):\n",
    "  layers = mapPanel.layers()\n",
    "  nLayer = layers.length()\n",
    "  for i in range(nLayer-1, 0, -1):\n",
    "    layer = layers.get(i)\n",
    "    if (layer.getName().match(name)) {\n",
    "      layers.remove(layer)\n",
    "    }\n",
    "\n",
    "\n",
    "\n",
    "  # reset the time series panel\n",
    "def resetTSPanel():\n",
    "  tsPanel.clear()\n",
    "  removeLayer('_')\n",
    "  removeLayer('Clicked')\n",
    "\n",
    "\n",
    "# ---------------------------------------------------------------\n",
    "# UIs:\n",
    "  # map panel\n",
    "mapPanel = ui.Map({'style': '{cursor': 'crosshair'}})\n",
    "mapPanel.centerObject(testArea, 10)\n",
    "mapPanel.setOptions('SATELLITE')\n",
    "mapPanel.addLayer(testArea, {'color': 'red'}, 'Test Area')\n",
    "mapPanel.addLayer(forestMask, mskVisParam, 'Forest Mask', False)\n",
    "\n",
    "  # menu panel\n",
    "trainButton = ui.Button('Train')\n",
    "ccdButton = ui.Button('Fit')\n",
    "nrtButton = ui.Button('Monitor')\n",
    "runButton = ui.Button('Run')\n",
    "saveButton = ui.Button('Save')\n",
    "menuSet = ui.Panel([ccdButton, nrtButton, runButton],\n",
    "                        ui.Panel.Layout.Flow('vertical'))\n",
    "ccdSelect = ui.Select(['Landsat', 'Sentinel-2', 'Sentinel-1'], 'Select sensor for CCD.', 'Landsat')\n",
    "LSTCheck = ui.Checkbox('Landsat', True)\n",
    "S2Check = ui.Checkbox('Sentinel-2', True)\n",
    "S1Check = ui.Checkbox('Sentinel-1', False)\n",
    "selectSet = ui.Panel([ccdSelect, LSTCheck, S2Check, S1Check],\n",
    "                        ui.Panel.Layout.Flow('vertical'))\n",
    "menuUISet = ui.Panel([menuSet, selectSet], ui.Panel.Layout.Flow('horizontal'))\n",
    "menuPanel = ui.Panel({\n",
    "  'widgets': [ui.Label('Menu'), menuUISet],\n",
    "  'layout': ui.Panel.Layout.Flow('vertical'),\n",
    "  'style': '{width': '20%'}})\n",
    "\n",
    "  # ts panel\n",
    "tsPanel = ui.Panel({\n",
    "  'widgets': [],\n",
    "  'style': '{position': 'bottom-right', 'width': '80%'}})\n",
    "\n",
    "  # ui panel\n",
    "controlPanel = ui.Panel({\n",
    "  'style': '{height': '30%'},\n",
    "  'widgets':[ui.SplitPanel(tsPanel, menuPanel, 'horizontal', False)]})\n",
    "mapPanel2 = ui.Panel({\n",
    "  'style': '{height': '70%'},\n",
    "  'widgets':[mapPanel]})\n",
    "uiPanel = ui.SplitPanel(mapPanel2, controlPanel, 'vertical')\n",
    "\n",
    "# ---------------------------------------------------------------\n",
    "# Runtime functions:\n",
    "trainButton.onClick(function() {\n",
    "  def saveCCD(ccd, sensor):\n",
    "    Export.image.toAsset({\n",
    "          'image': ccd,\n",
    "          'scale': 30,\n",
    "          'assetId': 'projects/bu-nearrealtime/lite/ccd/' + sensor + '_CCD',\n",
    "          'description': 'Save_' + sensor + '_CCD',\n",
    "          'region': testArea,\n",
    "          'maxPixels': 1e13,\n",
    "          'pyramidingPolicy': {'.default': 'sample'}\n",
    "    })\n",
    "  \n",
    "  LST = ut.getData(testArea, trainPeriod, 'Landsat')\n",
    "  saveCCD(ut.runCCD(LST, trainPeriod, 'NDFI').set({'region': 'test', 'sensor': 'Landsat'}), 'LST')\n",
    "  S2 = ut.getData(testArea, trainPeriod, 'Sentinel-2')\n",
    "  saveCCD(ut.runCCD(S2, trainPeriod, 'NDFI').set({'region': 'test', 'sensor': 'Sentinel-2'}), 'S2')\n",
    "  S1 = ut.getData(testArea, trainPeriod, 'Sentinel-1')\n",
    "  saveCCD(ut.runCCD(S1, trainPeriod, 'VV').set({'region': 'test', 'sensor': 'Sentinel-1'}), 'S1')\n",
    "})\n",
    "\n",
    "ccdButton.onClick(function() {\n",
    "  if (listener == 1) {\n",
    "    ccdButton.setLabel('Fit')\n",
    "    listener = 0\n",
    "  } else {\n",
    "    nrtButton.setLabel('Monitor')\n",
    "    ccdButton.setLabel('Cancel')\n",
    "    listener = 1\n",
    "  }\n",
    "})\n",
    "\n",
    "nrtButton.onClick(function() {\n",
    "  if (listener == 2) {\n",
    "    nrtButton.setLabel('Monitor')\n",
    "    listener = 0\n",
    "  } else {\n",
    "    nrtButton.setLabel('Cancel')\n",
    "    ccdButton.setLabel('Fit')\n",
    "    listener = 2\n",
    "  }\n",
    "})\n",
    "\n",
    "saveButton.onClick(function() {\n",
    "  Export.image.toAsset({\n",
    "        'image': alerts,\n",
    "        'scale': 30,\n",
    "        'description': 'SaveAlerts',\n",
    "        'assetId': 'Alerts',\n",
    "        'region': testArea,\n",
    "        'maxPixels': 1e13})\n",
    "})\n",
    "\n",
    "runButton.onClick(function() {\n",
    "  alerts = runNRT(testArea)\n",
    "  mapPanel.addLayer(alerts, altVisParam, 'Forest Disturbance Alerts')\n",
    "  saveButton.setDisabled(False)\n",
    "})\n",
    "\n",
    "mapPanel.onClick(function(coords) {\n",
    "  if (listener == 1) {\n",
    "    chartCCD(coords)\n",
    "  } else if (listener == 2) {\n",
    "    chartNRT(coords)\n",
    "  }\n",
    "})\n",
    "\n",
    "# ---------------------------------------------------------------\n",
    "# Initialization:\n",
    "#saveButton.setDisabled(True)\n",
    "ui.root.clear()\n",
    "ui.root.add(uiPanel)\n",
    "\n",
    "# End"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Display the interactive map"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "Map"
   ]
  }
 ],
 "metadata": {
  "anaconda-cloud": {},
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
